function [S_l,S_u,mq,kinvbs,wn] = mqconfreg_chen(u,p,alpha,varargin)
% This function computes the confidence region according to the approach by
% Chen et al (2017) as extended in Coblenz et al (2018) for bivariate data.
%
% call: [S_l,S_u,mq,kinv,wn] = mqconfreg_chen(u,p,alpha[,ploton])
%
% input     u                   - nx2 matrix of pseudo-observations
%           p                   - level at which multivariate quantile is 
%                                 computed; i.e., the 100%*p-quantile
%           alpha               - determines the confidence level 1-alpha;
%                                 i.e., for 90% confidence level enter 
%                                 alpha = 0.1
%           ploton (optional)   - controls whether a plot is produced or
%                                 not; default value: 1
%
% output    S_l                 - dlx2 matrix of coordinates of the lower
%                                 confidence region boundary
%           S_u                 - dux2 matrix of coordinates of the upper
%                                 confidence region boundary
%           boundary            - dx2 matrix of coordinates of the
%                                 multivariate quantile boundary
%           kinv                - estimated inverse of Kendall distribution
%                                 function
%           wn                  - estimated quantile of the Hausdorff
%                                 distance
%
% functions needed: ecopula, pobs
%
% Note that the precision of the estimation can be increased by: increasing
% the number of bootstraps n_bs, increasing the number of grid points n_ke 
% the kernel estimation is conducted on, or by implementing a more
% sophisticated method to compute the Hausdorff distance. The suggested
% settings reflect a tested tradeoff between computation time and 
% precision.
%
% References:
% Chen, Genovese, Wasserman (2017), Density Level Sets: Asymptotics, 
% Inference, and Visualization. Journal of the American Statistical 
% Association, 112(520): 16841696.
% Coblenz, Dyckerhoff, Grothe (2018), Confidence Regions for Multivariate 
% Quantiles, Working Paper.
%
% updated: 24.05.2018
%
% Author: Maximilian Coblenz, 2018
% This code is released under the 3-clause BSD license.
% 

% some parsing and sanity checks
ip = inputParser;
ip.addRequired('u',@ismatrix);
ip.addRequired('p',@isscalar);
ip.addRequired('alpha',@isscalar);
ip.addOptional('ploton',@isscalar);
ip.parse(u,p,alpha,varargin{:});

if (min(u(:)) < 0 || max(u(:)) > 1) % rank-transform data
    
    u = pobs(u);
    
end

if (p < 0 || p > 1)
    error('the quantile level must be between 0 and 1');
end

if (alpha < 0 || alpha > 1)
    error('the confidence level must be between 0 and 1');
end

% initialization of variables
if nargin < 4    
    ploton = 1;   
else    
    ploton = varargin{1};   
end

n_bs = 1000; % number of bootstrap repetitions
n_ke = 100; % number of kernel estimation grid points
n = size(u,1);
Wstar = zeros(1,n_bs);
h = 1/nthroot(n,6); % Silverman's rule of thumb
bootstrap_indices = randi(size(u,1),n,n_bs); % bootstrap indices

% compute invers Kendall distribution function and multivariate quantile
p_grid = (1/size(u,1):1/size(u,1):1);
V_I = (ecopula(u,u)*(size(u,1)+1)-1)/(size(u,1)-1);
kendallfunc = sum(V_I*ones(1,length(p_grid))<=ones(size(u,1),1)*p_grid,1)/(size(u,1));
[~,b] = min(kendallfunc(:)+(kendallfunc(:)<p)*10000000);
kinv = p_grid(b);

% generate kernel estimate for cdf of transformed data
data_trans = norminv(u,0,1);
gridx1 = linspace(min(data_trans(:,1))-1,max(data_trans(:,1))+1,n_ke);
gridx2 = linspace(min(data_trans(:,2))-1,max(data_trans(:,2))+1,n_ke);
[x1,x2] = meshgrid(gridx1, gridx2);
x1 = x1(:);
x2 = x2(:);
xi = [x1 x2];
[fke,~] = ksdensity(data_trans,xi,'Function','cdf','Bandwidth',h);
Z = reshape(fke,n_ke,n_ke);

C = contourc(gridx1,gridx2,Z,[kinv kinv]);
C = C(:,2:end)';
mq = normcdf(C,0,sqrt(1+h^2));
mq = [[mq(1,1) 1]; mq; [1 mq(end,2)]];

% now bootstrap to get Wstar
for ii = 1:1:n_bs
    
    % bootstrap data and get K^-1 of bootstrapped data
    databs = u(bootstrap_indices(:,ii),:);
    p_grid = (1/size(databs,1):1/size(databs,1):1);
    V_I = (ecopula(databs,databs)*(size(databs,1)+1)-1)/(size(databs,1)-1);
    kendallfunc = sum(V_I*ones(1,length(p_grid))<=ones(size(databs,1),1)*p_grid,1)/(size(databs,1));
    [~,b] = min(kendallfunc(:)+(kendallfunc(:)<p)*10000000);
    kinvbs = p_grid(b);
    
    % generate ksdensity of transformed bs data
    data_trans = norminv(databs,0,1);
    gridx1 = linspace(min(data_trans(:,1))-1,max(data_trans(:,1))+1,n_ke);
    gridx2 = linspace(min(data_trans(:,2))-1,max(data_trans(:,2))+1,n_ke);
    [x1,x2] = meshgrid(gridx1, gridx2);
    x1 = x1(:);
    x2 = x2(:);
    xi = [x1 x2];
    [fke,~] = ksdensity(data_trans,xi,'Function','cdf','Bandwidth',h);
    Z = reshape(fke,n_ke,n_ke);
    
    C = contourc(gridx1,gridx2,Z,[kinvbs kinvbs]);
    C = C(:,2:end)';
    Cu_bs = normcdf(C,0,sqrt(1+h^2));
    if isempty(Cu_bs)
        Cubs = [1 1];
    else
        Cubs = [[Cu_bs(1,1) 1]; Cu_bs; [1 Cu_bs(end,2)]];
    end
    
    % now compute Hausdorff distance between Cu and Cubs
    delta_h = max([abs(mq(1,1)-Cubs(1,1)) abs(mq(end,2)-Cubs(end,2))]);
    
    % pointwise distance, this is a simplified version of the Hausdorff
    % distance
    for jj = 1:1:size(Cubs,1)
        
        d = min(vecnorm(mq - Cubs(jj,:),2,2));
        delta_h = max([delta_h d]);
        
    end % jj
    
    Wstar(ii) = delta_h;
    
end % ii

% compute wn
wn = quantile(Wstar,1-alpha);

% compute confidence region boundaries S_l and S_u
point = 0.5*(mq(1:end-1,:)+mq(2:end,:));
direc = mq(1:end-1,:)-mq(2:end,:);
direc = [direc(:,2) -direc(:,1)];
S_l = point - direc./(vecnorm(direc,2,2))*wn;
S_u = point + direc./(vecnorm(direc,2,2))*wn;

% plot
if ploton
    
    figure
    plot(mq(:,1),mq(:,2),'-b')
    hold on
    plot(S_l(:,1),S_l(:,2),'-','color',[1 0.6 0]);
    plot(S_u(:,1),S_u(:,2),'-','color',[1 0.6 0]);
    plot(u(:,1),u(:,2),'.k')
    hold off
    xlim([0 1])
    ylim([0 1])
    xlabel('U_1')
    ylabel('U_2')
    
end


end